"""Basic building block of the Entitled library.

Rules are essentially wrappers around boolean functions that determines a relationship
between an actor and a resourc.

Resource can essentially be any entity in an application.
Actors represents any entity that, in an application, can act upon a resource.
"""

import typing

from entitled import exceptions

ResourceModel = typing.TypeVar("ResourceModel", contravariant=True)
ActorModel = typing.TypeVar("ActorModel", contravariant=True)


class Rule(typing.Generic[ActorModel, ResourceModel]):
    """Base class for rules

    Args:
        name (str): The name given to this rule.
        rule (Callable): The underlying boolean function.
    """

    def __init__(
        self, name: str, rule: typing.Callable[[ActorModel, ResourceModel, dict], bool]
    ):
        self.name = name
        self.rule = rule

    def __call__(
        self,
        actor: ActorModel,
        resource: ResourceModel,
        context: typing.Optional[dict] = None,
    ) -> bool:
        if not context:
            context = {}
        return self.rule(actor, resource, context)

    def __and__(self, rh_rule: "Rule[ActorModel, ResourceModel]"):
        return Rule(
            f"{self.name}&{rh_rule.name}",
            lambda actor, resource, context: (
                self(actor, resource, context) and rh_rule(actor, resource, context)
            ),
        )

    def __or__(self, rh_rule: "Rule[ActorModel, ResourceModel]"):
        return Rule(
            f"{self.name}|{rh_rule.name}",
            lambda actor, resource, context: (
                self(actor, resource, context) or rh_rule(actor, resource, context)
            ),
        )

    def authorize(
        self,
        actor: ActorModel,
        resource: ResourceModel,
        context: typing.Optional[dict] = None,
    ) -> bool:
        if not self(actor, resource, context):
            raise exceptions.AuthorizationException("Unauthorized")
        return True



"""Centralization of all decision-making processes on a single decision point"""

import importlib.util
import pathlib
import types
import typing

from entitled import policies


class Client:
    "The Client class for decision-making centralization."

    def __init__(self, base_path="policies"):
        self._policy_registrar: dict[typing.Type, policies.Policy] = {}
        self._load_path = pathlib.Path(base_path)
        self.load_policies_from_path(self._load_path)

    def authorize(self, actor, action, resource, context: typing.Optional[dict] = None):
        policy = self._policy_lookup(resource)
        return policy.authorize(actor, action, resource, context)

    def allows(
        self, actor, action, resource, context: typing.Optional[dict] = None
    ) -> bool:
        policy = self._policy_lookup(resource)
        return policy.allows(actor, action, resource, context)

    def actions(self, actor, resource, context: typing.Optional[dict] = None):
        policy = self._policy_lookup(resource)
        return policy.actions(actor, resource, context)

    def register_policy(self, policy: policies.Policy):
        if hasattr(policy, "__orig_class__"):
            resource_type = getattr(policy, "__orig_class__").__args__[0]
            if resource_type not in self._policy_registrar:
                self._policy_registrar[resource_type] = policy
            else:
                raise ValueError(
                    "A policy is already registered for this resource type"
                )
        else:
            raise AttributeError(f"Policy {policy} is incorrectly defined")

    def reload_registrar(self):
        self.load_policies_from_path(self._load_path)

    def load_policies_from_path(self, path: pathlib.Path):
        for file_path in path.glob("*.py"):
            mod_name = file_path.stem
            full_module_name = ".".join(file_path.parts[:-1] + (mod_name,))
            spec = importlib.util.spec_from_file_location(full_module_name, file_path)
            if spec:
                module = importlib.util.module_from_spec(spec)
                if spec.loader:
                    try:
                        spec.loader.exec_module(module)
                    except Exception as e:
                        raise e

                    self._register_from_module(module)

    def _register_from_module(self, module: types.ModuleType):
        for attribute_name in dir(module):
            attr = getattr(module, attribute_name)
            if isinstance(attr, policies.Policy):
                try:
                    self.register_policy(attr)
                except (ValueError, AttributeError):
                    pass

    def _policy_lookup(self, resource) -> policies.Policy:
        if type(resource) not in self._policy_registrar:
            raise ValueError("No policy registered for this resource type")

        return self._policy_registrar[type(resource)]



"""Grouping of authorization rules around a particular resource type"""

import typing

from entitled import exceptions
from entitled.rules import Rule

T = typing.TypeVar("T")


class Policy(typing.Generic[T]):
    """A grouping of rules refering the given resource type."""

    def __init__(
        self,
        rules: typing.Optional[dict[str, list[Rule[typing.Any, T]]]] = None,
    ):
        self._registry: dict[
            str,
            list[Rule[typing.Any, T]],
        ] = {}

        if not rules:
            rules = {}

        for action, rule in rules.items():
            self.register(action, *rule)

    def register(self, action, *rules: Rule[typing.Any, T]):
        if action in self._registry:
            self._registry[action].append(*rules)
        else:
            self._registry[action] = [*rules]

    def actions(self, actor, resource: T, context: typing.Optional[dict] = None):

        return filter(
            lambda action: self.allows(actor, action, resource, context),
            self._registry.keys(),
        )

    def allows(
        self,
        actor,
        action,
        resource: T,
        context: typing.Optional[dict] = None,
    ) -> bool:
        try:
            return self.authorize(actor, action, resource, context)
        except exceptions.AuthorizationException:
            return False

    def authorize(
        self,
        actor,
        action,
        resource: T,
        context: typing.Optional[dict] = None,
    ) -> bool:
        if action not in self._registry:
            raise exceptions.UndefinedAction(
                f"Action <{action}> undefined for this policy"
            )

        if not any(rule(actor, resource, context) for rule in self._registry[action]):
            raise exceptions.AuthorizationException("Unauthorized")

        return True
